# See LICENSE.md file in project root directory

import os
import json
import shutil
import signal
import traceback
from datetime import datetime
from retry_requests import retry

from .Stax import Stax

sess = retry()

def log(*args):
    print("[" + str(datetime.now()) + "]", *args)

def result(obj):
    with open('response.json', 'w') as f:
        json.dump(obj, f)

# Define module wrapper decorator
def def_module(module_id:str, module_key:str):
    def wrapper(app):
        def inner(*args, **kwargs):
            TEST = "test" in kwargs and kwargs["test"]

            # Setup keyboard interrupt handler (if worker is killed)
            signal.signal(signal.SIGINT, signal.default_int_handler)

            # Load request.json file
            if not os.path.exists('request.json'):
                raise Exception("Missing request.json file.")

            with open('request.json') as f:
                req = json.load(f)

            try:
                job:str = req["jobId"]
                team:str = req["teamId"]
                document:str = req["documentId"]
                stack:str = req["stackId"]
                step:str = req["stepId"]
                stepN:str = req["step"]
                config:list[dict] = req["config"]

            except:
                raise Exception("Missing required fields in request.json.")

            # Step identifier to report status to the Stax.ai API
            report = {
                "step": stepN,
                "stepId": step,
                "stack": stack,
            }

            # Instantiate a Stax.ai API object
            stax = Stax(
                module_id=module_id,
                module_key=module_key,
                team_id=team,
                document=document,
                api_url=os.getenv('STAX_API_URL', default='https://api.stax.ai')
            )

            try:
                # Create tmp working directory
                if not os.path.exists('./tmp'):
                    os.mkdir('./tmp')

                log("Started job:", job)
                
                # Call module/app function
                count = app(stax, document, stack, team, config)

                # Finished job, post result to Stax.ai API
                result({ "status": "Complete", "units": count })
                if not TEST:
                    stax.post('/job/complete/' + job, {
                        **report,
                        "count": count
                    })

                log("Completed job:", job)

            except KeyboardInterrupt:
                log("SIGINT Received. Stopping job:", job)
                result({ "status": "Killed" })

                # Let the Stax.ai API know what happened so it can handle it
                if not TEST:
                    stax.post('/job/complete/' + job, {
                        **report,
                        "error": "Automate module worker was stopped. This job should be retried shortly."
                    })

                exit()

            except Exception as e:
                log("Error in job:", job)
                trace = traceback.format_exc()
                print(trace)

                error = {
                    "error": str(e),
                    "traceback": str(trace)
                }
                result({ "status": "Error", **error })

                # Report error to the Stax.ai API
                if not TEST:
                    stax.post('/job/complete/' + job, {
                        **report,
                        **error
                    })

            finally:
                # Delete tmp working directory
                if os.path.exists('./tmp'):
                    shutil.rmtree('./tmp')

        return inner
    return wrapper


# See LICENSE.md file in project root directory

def getConfigByLabel(config:'list[dict]', label:str):
    for c in config:
        if "label" not in c or "value" not in c:
            continue
        if c["label"] == label:
            return c["value"]
    return None


def getValueByKey(fields:'list[dict]', key:str) -> (str):
    for f in fields:
        if "key" in f and "value" in f and f["key"] == key:
            return f["value"]
    return ""


def getFieldByKey(fields:'list[dict]', key:str):
    for f in fields:
        if "key" not in f or "value" not in f:
            continue
        if f["key"] == key:
            return f
    return None


def getFieldsByKey(fields:'list[dict]', key:str):
    ret = []
    for f in fields:
        if "key" not in f or "value" not in f:
            continue
        if f["key"] == key:
            ret.append(f)
    return ret


def getFieldIndex(fields:'list[dict]', key:str, value:str):
    for i,f in enumerate(fields):
        if "key" not in f or "value" not in f:
            continue
        if f["key"] == key and f["value"] == value:
            return i
    return None


# See LICENSE.md file in project root directory
import json
from retry_requests import retry
from .helpers import getConfigByLabel

class Stax:
    def __init__(self, module_id:str, module_key:str, team_id:str=None, document:str=None, api_url:str="https://api.stax.ai"):
        # Construct API header
        self.headers = {
            'Module': module_id,
            'Authorization': 'Bearer ' + module_key,
            'Team': team_id
        }

        self.document_id = document
        self.api_url = api_url

        # Create retry requests session
        self.sess = retry()


    def getDocument(self, presigned:bool=True, document:str=None):
        '''
        Get document metadata.
        '''
        document_id = document if document else self.document_id
        if not document_id:
            raise Exception("No document ID specified")

        # Construct endpoint
        endpoint = '/document/get?docId=' + document_id
        if presigned:
            endpoint += '&presigned=true'

        res = self.get(endpoint)
        return res['doc']


    def downloadDocument(self, document:str=None, path:str=None):
        '''
        Download raw document.
        If path is not specified, this function returns the binary data.
        '''
        document_id = document if document else self.document_id
        if not document_id:
            raise Exception("No document ID specified")

        res = self.get('/document/download?docId=' + document_id, raw=True)
        if path is None:
            return res
        
        with open(path, 'wb') as f:
            f.write(res.content)


    def downloadPage(self, key:str=None, document:str=None, path:str=None):
        '''
        Download a single document page - make sure you pass the page key.
        If path is not specified, this function returns the binary data.
        '''
        document_id = document if document else self.document_id
        if not key:
            raise Exception("Missing page key")
        if not document_id:
            raise Exception("No document ID specified")

        res = self.get('/document/downloadPage?docId=' + document_id + '&pagePath=' + key, raw=True)
        if path is None:
            return res
        
        with open(path, 'wb') as f:
            f.write(res.content)

    
    def updateDocument(self, diff={}, document:str=None):
        '''
        Update a document with a provided diff of changes.
        '''
        document_id = document if document else self.document_id
        if not document_id:
            raise Exception("No document ID specified")
        
        self.post('/document/update', {
            "docId": document_id,
            "docDiff": diff
        })


    def downloadConfigFile(self, config:'list[dict]'=[], label:str=None, path:str=None):
        '''
        Download config file by label
        '''
        if not label:
            raise Exception("Missing configuration label")

        src = getConfigByLabel(config, label)
        if not src:
            raise Exception("Configuration parameter does not exist.")

        # TODO - download


    def get(self, endpoint:str, raw:bool=False):
        '''
        Make a GET request from the Stax.ai API.
        Set 'raw' to True to return raw API response.
        '''
        res = self.sess.get(self.api_url + endpoint, headers=self.headers)
        return res if raw else self._handle_response_(res)


    def post(self, endpoint:str, data={}, raw:bool=False):
        '''
        Make a POST request to the Stax.ai API with provided JSON data.
        Set 'raw' to True to return raw API response.
        '''
        res = self.sess.post(self.api_url + endpoint, json=data, headers=self.headers)
        return res if raw else self._handle_response_(res)


    def _handle_response_(self, res):
        if res.status_code != 200:
            raise Exception("Failed to communicate with Stax.ai API")
        try:
            data = res.json()
            if not data['success']:
                raise Exception("API Error:" + json.dumps(data))

            return data
        except Exception as e:
            raise Exception("API Parse Error:" + str(e) + "\n" + res.content)



# See LICENSE.md file in project root directory
import re
from typing import Dict
from .helpers import getValueByKey, getConfigByLabel

def renderPyText(doc: Dict, body: str):
    out = body
    
    # Link to document
    link = "https://my.stax.ai/stack/" + (str(doc["stack"]) if doc["stack"] else "unclassified") + "/document/" + str(doc["_id"])
    
    # Field value extractor
    def field(key):
        return getValueByKey(doc["metadata"], key)
    
    for m in re.finditer(r"\{(.*?)\}", body, re.MULTILINE | re.DOTALL):
        for g in range(0, len(m.groups())):
            pattern = m.group(g)
            try:
                value = str(eval(pattern[1:-1]))
            except Exception as e:
                raise Exception("Failed to evaluate expression: " + pattern + " - " + str(e))
                
            out = out.replace(pattern, value)
    
    return out


from setuptools import setup, find_packages

with open('README.md', 'r') as f:
    readme = f.read()

setup(
    name='stax_module_sdk',
    version='0.1.5',
    url='https://bitbucket.org/pinetree-ai/stax-module-sdk-py/src/master/',
    author='Stax.ai, Inc. <https://stax.ai>',
    author_email='naru@stax.ai',
    description='Stax.ai Module SDK',
    long_description=readme,
    long_description_content_type='text/markdown',
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent"
    ],
    license='CC-BY-NC 4.0',
    python_requires='>=3.6',
    package_dir={'': 'src'},
    packages=find_packages(where='src'),
    install_requires=['retry_requests'],
    scripts=['bin/stax_module_cli']
)